print("import: 'tornado'")
import tornado

print("import: 'tornado.platform'")
import tornado.platform

print("import: 'tornado.test'")
import tornado.test

print("import: 'tornado.gen'")
import tornado.gen

print("import: 'tornado.netutil'")
import tornado.netutil

